pub struct MonoidalString<C, M: ?Sized> { /* private fields */ }Expand description
Creates free-arithmetic constructions based upon free-multiplication of letters of type C
that uses a Vec<C> internally
Basic Construction
Given type parameters C and M, the construction goes as follows:
- Internally, each instance contains a list of values of type
C - By default, multiplication is given by concatenating the internal
Vec’s of each argument - Then, with the above as a base, the given type
Mcan modify that operation by applying a multiplication rule to the internal list every time anotherCis appended to it
With this system, using different M rules and C types, a number of variants can be defined:
FreeMonoid<C>is created by using the default multiplication with anyCFreeGroup<C>is made by wrapping it in aFreeInv<C>to allow eachCto be inverted and by and modifyingMto make inverses cancelFreePowMonoid<C,P>is constructed by wrapping eachCin aFreePow<C>to raise each element to a symbolic powerPand by havingMcombine any adjacent elements with equal base by adding their exponents
Other impls
In addition to the basic multiplication traits, MonoidalString implements a number of other notable traits depending on the type arguments:
- Div, DivAssign, Inv, etc. These apply whenever
Mimplements - Index wraps the internal list’s implementation
InvMonoidRule<C>to provide a way to invertCelements - MulAssociative and MulCommutative whenever
Mimplements AssociativeMonoidRule or CommutativeMonoidRule - PartialEq with any type that implements
Borrow<[C]>. This is in order to make it possible to test equality with other structures (likeVec<C>and[C]) that are lists ofC’s - PartialOrd and Ord implement lexicographic ordering
- Extend, FromIterator, and Product all are implemented by applying the multiplication
rule to an Iterator of
C’s.
A note on IterMut
The method [MonoidalString.iter_mut()] will give an valid iterator over mutable references to each
element of the object, but it is of note that because of the nature of some of the MonoidRule’s,
the method will reallocate internal storage and iterate through every element,
even if dropped.
The reason for this is that if it did not, it would be possible to modify a MonoidalString into an invalid state. Hence, to mitigate this, the iterator must remultiply every element again as it iterates over the references.
Implementations
Sourceimpl<C, M: ?Sized> MonoidalString<C, M>
impl<C, M: ?Sized> MonoidalString<C, M>
Sourcepub fn len(&self) -> usize
pub fn len(&self) -> usize
Returns the number of letters in this monoidal-string
Examples
use maths_traits::algebra::One;
use free_algebra::FreeMonoid;
let x = FreeMonoid::one();
let y = x.clone() * 'a' * 'a';
assert_eq!(x.len(), 0);
assert_eq!(y.len(), 2);
Sourcepub fn iter(&self) -> Iter<'_, C>
pub fn iter(&self) -> Iter<'_, C>
Produces an iterator over references to the letters in this element
Examples
use maths_traits::algebra::One;
use free_algebra::FreeMonoid;
let x = FreeMonoid::one() * 'a' * 'b' * 'c';
let mut iter = x.iter();
assert_eq!(iter.next(), Some(&'a'));
assert_eq!(iter.next(), Some(&'b'));
assert_eq!(iter.next(), Some(&'c'));
assert_eq!(iter.next(), None);
Sourcepub fn iter_mut(&mut self) -> IterMut<'_, C, M> ⓘwhere
M: MonoidRule<C>,
pub fn iter_mut(&mut self) -> IterMut<'_, C, M> ⓘwhere
M: MonoidRule<C>,
Produces an iterator over mutable references to the letters in this element
Note that the potential for modification does mean that the element needs to be remultiplied as letters are changed, so a potential reallocation of space may occur.
Examples
use free_algebra::{FreePow, FreePowMonoid};
let x = FreePow('a',1) * FreePow('b',1) * FreePow('c',1);
let mut y = x.clone();
for FreePow(c,p) in y.iter_mut() {
*c = 'a';
}
assert_eq!(x, [FreePow('a',1), FreePow('b',1), FreePow('c',1)]);
assert_eq!(y, [FreePow('a',3)]);
Sourcepub fn reverse(self) -> Selfwhere
Self: Product<C>,
pub fn reverse(self) -> Selfwhere
Self: Product<C>,
Reverses the letters in this element and remultiplies
Examples
use maths_traits::algebra::One;
use free_algebra::FreeMonoid;
let x = FreeMonoid::one() * 'a' * 'b' * 'c';
let y = x.clone().reverse();
assert_eq!(x, ['a', 'b', 'c']);
assert_eq!(y, ['c', 'b', 'a']);
Sourcepub fn commutator(self, rhs: Self) -> Self
pub fn commutator(self, rhs: Self) -> Self
Computes the multiplicative commutator [a,b] = a⁻¹b⁻¹ab
Examples
use free_algebra::FreeGroup;
use free_algebra::FreeInv::*;
let x:FreeGroup<_> = Id('a').into();
let y:FreeGroup<_> = Id('b').into();
assert_eq!(x, [Id('a')]);
assert_eq!(y, [Id('b')]);
assert_eq!(x.commutator(y), [Inv('a'), Inv('b'), Id('a'), Id('b')]);
A property of significance for this product is that when the arguments commute, the
output is always 1.
use maths_traits::algebra::One;
use free_algebra::{FreePowMonoid, FreePow};
let x:FreePowMonoid<_,_> = FreePow('a', 2).into();
let y:FreePowMonoid<_,_> = FreePow('a', -24).into();
assert!(x.commutator(y).is_one());